
/*
 * Copyright (c) 2002-2008, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the MVAPICH software package, developed by the
 * team members of The Ohio State University's Network-Based Computing
 * Laboratory (NBCL) and headed by Professor Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licensing information, please refer to the
 * copright file COPYRIGHT_MVAPICH in the top level MPICH directory.
 *
 */


#if ((defined(VIADEV_RPUT_SUPPORT) || defined(VIADEV_RGET_SUPPORT)))

#include "mpiimpl.h"
#include "mpimem.h"
#include "mpid.h"
#include "mpiops.h"
#include "mpipt2pt.h"

/* For device level initialization */
#ifdef USE_MRAIL
#include "../../mpid/vapi_multirail/collutils.h"
#elif defined(CH_GEN2)
#include "../../mpid/ch_gen2/collutils.h"
reg_entry* coll_reg;
#elif (!defined(CH_GEN2_MRAIL))
#include "../../mpid/vapi/collutils.h"
#endif

/* Prototype for communicator RDMA features initialization*/
void comm_rdma_init(struct MPIR_COMMUNICATOR*);
void comm_exch_addr(struct MPIR_COMMUNICATOR*);
void set_local_offsets(struct MPIR_COMMUNICATOR* comm);
void coll_prepare_flags(struct Collbuf*);
#ifdef _SMP_
extern int disable_shmem_barrier;
#endif
void comm_rdma_init(struct MPIR_COMMUNICATOR* comm)
{
#ifndef CH_GEN2_MRAIL
    struct L_Collbuf* local_coll;
    struct R_Collbuf* remote_coll;

    if(comm->np <= 1) {
        /* We do not need to do any initialization */

        comm->collbuf = NULL;
#if ((!defined(CH_GEN2_MRAIL)) && (!defined(USE_MRAIL)) )        
        comm->is_allgather_enabled = 0;
#endif
        comm->is_alltoall_enabled = 0;
        comm->is_mcast_enabled = 0;
        comm->is_barrier_enabled = 0;
        return;
    }

    comm->collbuf = NULL;

    char* value;
    comm->is_alltoall_enabled = 0;
    if ((value = getenv("VIADEV_USE_RDMA_ALLTOALL")) != NULL) {
        comm->is_alltoall_enabled = atoi(value);
    }

#if ((!defined(CH_GEN2_MRAIL)) && (!defined(USE_MRAIL)) )

    comm->is_allgather_enabled = 0;
    if ((value = getenv("VIADEV_USE_RDMA_ALLGATHER")) != NULL) {
        comm->is_allgather_enabled = atoi(value);
    }
#endif
    
    if(NULL != getenv("DISABLE_HARDWARE_MCST")) {
        /* The user has disabled Multicast */
        comm->is_mcast_enabled = 0;
    } else {
        /* The user has enabled Multicast */
        comm->is_mcast_enabled = 1;
    }

    comm->is_barrier_enabled = 0;
    if ((value = getenv("VIADEV_USE_RDMA_BARRIER")) != NULL) {
        comm->is_barrier_enabled = atoi(value);
    }

#ifdef _SMP_
    if ((disable_shmem_barrier == 0) && (comm->is_barrier_enabled)){
        
        printf("Only VIADEV_USE_SHMEM_BARRIER or VIADEV_USE_RDMA_BARRIER can be 1 at one time\n");
        MPI_Abort(comm->self,0);
    }
#endif

    comm->collbuf = (struct Collbuf*)
        malloc(sizeof(struct Collbuf));

    CHECK_ALLOC(comm->collbuf);

    local_coll = (struct L_Collbuf*)
        malloc(sizeof(struct L_Collbuf));

    CHECK_ALLOC(local_coll);

    local_coll->buf = (void*) malloc(COLL_BUF_SIZE);

    CHECK_ALLOC(local_coll->buf);

    /* Set the collective buffer to zero */
    memset(local_coll->buf, 0, COLL_BUF_SIZE);

    /* Register the local memory */
#ifdef CH_GEN2
    local_coll->mr = register_memory(local_coll->buf, COLL_BUF_SIZE);
    
    coll_reg = (reg_entry*) malloc(sizeof(reg_entry));    
    coll_reg->buf = local_coll->buf;
    coll_reg->valid = 1;
    coll_reg->mem_hndl = local_coll->mr;

    if (viadev.coll_comm_reg == NULL){
        coll_reg->next = NULL;
        viadev.coll_comm_reg = coll_reg;
    }
    else{
        coll_reg->next = viadev.coll_comm_reg;
        viadev.coll_comm_reg = coll_reg;
    }

#else
    register_mem(local_coll->buf, COLL_BUF_SIZE,
            &(local_coll->buf_hndl));
#endif

    comm->collbuf->l_coll = local_coll;

    /* Allocate space for
     * storing remote buffer
     * addresses */
    remote_coll = (struct R_Collbuf*)
        malloc(sizeof(struct R_Collbuf) * comm->np);

    CHECK_ALLOC(remote_coll);

    /* Set all the memory to zero */
    memset(remote_coll, 0, sizeof(struct R_Collbuf) * comm->np);

    comm->collbuf->r_coll = remote_coll;

    /*Set the flags for rdma barrier */
    comm->collbuf->togle = 0;
    comm->collbuf->rdma_barrier_id = 0;
#endif
}

/* Exchange addresses with other processes */
void comm_exch_addr(struct MPIR_COMMUNICATOR* comm)
{
#ifndef CH_GEN2_MRAIL    
    int i = 0, right, left;
    struct Coll_Addr_Exch send_pkt;
    struct Coll_Addr_Exch *recv_pkt;

    MPI_Status* statarray = NULL;
    MPI_Request* reqarray = NULL;

    if(comm->np <= 1) {
        return;
    }

    recv_pkt = (struct Coll_Addr_Exch*)
        malloc(sizeof(struct Coll_Addr_Exch) * comm->np);
    CHECK_ALLOC(recv_pkt);

    statarray = (MPI_Status*) malloc(2 * sizeof(MPI_Status) * comm->np);
    CHECK_ALLOC(statarray);

    reqarray = (MPI_Request*) malloc(2 * sizeof(MPI_Request) * comm->np);
    CHECK_ALLOC(reqarray);

    send_pkt.buf = comm->collbuf->l_coll->buf;
#ifdef CH_GEN2
    send_pkt.rkey = comm->collbuf->l_coll->mr->rkey;
#else
    send_pkt.buf_hndl = comm->collbuf->l_coll->buf_hndl;
#endif

    right = (comm->local_rank + 1) % comm->np;
    left = (comm->local_rank + comm->np - 1) % comm->np;

    for(i = 0; i < comm->np - 1; i++) {

        MPI_Sendrecv((void*)&send_pkt, sizeof(struct Coll_Addr_Exch),
                MPI_BYTE, comm->lrank_to_grank[right], ADDR_EXCHANGE_TAG,
                (void*)&(recv_pkt[left]),sizeof(struct Coll_Addr_Exch),
                MPI_BYTE, comm->lrank_to_grank[left], ADDR_EXCHANGE_TAG,
                MPI_COMM_WORLD, &(statarray[i]));

        right = (right+1) % comm->np;
        left = (left + comm->np - 1) % comm->np;
    }

    for(i = 0; i < comm->np; i++) {
        comm->collbuf->r_coll[i].buf = recv_pkt[i].buf;
#ifdef CH_GEN2
        comm->collbuf->r_coll[i].rkey = recv_pkt[i].rkey;
#else
        comm->collbuf->r_coll[i].buf_hndl.hndl = recv_pkt[i].buf_hndl.hndl;
        comm->collbuf->r_coll[i].buf_hndl.lkey = recv_pkt[i].buf_hndl.lkey;
        comm->collbuf->r_coll[i].buf_hndl.rkey = recv_pkt[i].buf_hndl.rkey;
#endif
    }

    free(reqarray);
    free(statarray);
    free(recv_pkt);
#endif
}

void set_local_offsets(struct MPIR_COMMUNICATOR* comm)
{
#ifndef CH_GEN2_MRAIL
    struct L_Collbuf* l_coll;
    struct R_Collbuf* r_coll;
    struct Alltoall* atoa = NULL;
    struct Direct_Eager* de = NULL;
    struct DE_Remote_Node* rnode = NULL;
    int i;

    if(comm->np <= 1) {
        return;
    }

    l_coll = comm->collbuf->l_coll;
    r_coll = comm->collbuf->r_coll;

    atoa = (struct Alltoall*) malloc(sizeof(struct Alltoall));
    CHECK_ALLOC(atoa);


    /* Direct Eager stuff */
    de = (struct Direct_Eager*) malloc(sizeof(struct Direct_Eager));
    CHECK_ALLOC(de);

    rnode = (struct DE_Remote_Node*) malloc(sizeof(struct DE_Remote_Node) *
            comm->np);
    CHECK_ALLOC(rnode);

    for(i = 0; i < comm->np; i++) {
        /* Don't set for myself */
        if ( comm->local_rank == i ) continue;

#if defined(CH_GEN2)
        rnode[i].recv_addr[0] = (void*)
            ((char*)l_coll->buf + COLL_FLAG_SPACE +
             COLL_ADDR_SEND_SPACE +
             (i* (sizeof(aint_t) + sizeof(uint32_t) + sizeof(char))));
#else
        rnode[i].recv_addr[0] = (void*)
            ((char*)l_coll->buf + COLL_FLAG_SPACE +
             COLL_ADDR_SEND_SPACE +
             (i* (sizeof(aint_t) + sizeof(VAPI_rkey_t) + sizeof(char))));
#endif

        rnode[i].recv_addr[1] = (void*)
            ((char*)rnode[i].recv_addr[0] +
             (COLL_ADDR_RECV_SPACE/2));

        rnode[i].done[0] = (void*)
            ((char*)l_coll->buf + COLL_FLAG_SPACE +
             COLL_ADDR_SEND_SPACE + COLL_ADDR_RECV_SPACE +
             i*sizeof(char));

        rnode[i].done[1] = (void*)
            ((char*)rnode[i].done[0] + (COLL_ACK_SPACE/2));

#if defined(CH_GEN2)
        rnode[i].dest_addr[0] = (void*)
            ((char*)r_coll[i].buf + COLL_FLAG_SPACE +
             COLL_ADDR_SEND_SPACE +
             (comm->local_rank*(sizeof(aint_t) + sizeof(uint32_t)
                                + sizeof(char))));
#else
        rnode[i].dest_addr[0] = (void*)
            ((char*)r_coll[i].buf + COLL_FLAG_SPACE +
             COLL_ADDR_SEND_SPACE +
             (comm->local_rank*(sizeof(aint_t) + sizeof(VAPI_rkey_t)
                                + sizeof(char))));
#endif

        rnode[i].dest_addr[1] = (void*)
            ((char*)rnode[i].dest_addr[0] +
             (COLL_ADDR_RECV_SPACE/2));

        rnode[i].r_done[0] = (void*)
            ((char*)r_coll[i].buf + COLL_FLAG_SPACE +
             COLL_ADDR_SEND_SPACE + COLL_ADDR_RECV_SPACE +
             (comm->local_rank*sizeof(char)));

        rnode[i].r_done[1] = (void*)
            ((char*)rnode[i].r_done[0] + (COLL_ACK_SPACE/2));

    }

    /* Start with the first set of buffers */
    de->alternate = 0;

    de->node = rnode;

    atoa->de = de;

    comm->collbuf->atoa = atoa;

    /*temp*/
    /* Set the Collective RDMA flag */
    coll_prepare_flags(comm->collbuf);
#endif
}

/* A wrap around function to prepare the
 * flags in the beginning of the collective
 * buffer
 */
void coll_prepare_flags(struct Collbuf* coll)
{
#ifndef CH_GEN2_MRAIL
    /* Set the first byte to RDMA_FLAG */

    *((char*)coll->l_coll->buf) = ALLTOALL_RDMA_FLAG;

    coll->atoa->rdma_flag = coll->l_coll->buf;

    /* Only a 1 byte long field */
    coll->atoa->rdma_flag_size = 1;
#endif
}

#if ((!defined(CH_GEN2_MRAIL)) && (!defined(USE_MRAIL)) )
void comm_rdma_allgather_init(struct MPIR_COMMUNICATOR *comm)
{
	struct Allgather *allgather;
	int dimensions, i, j, mask, dest;

	struct L_Collbuf *l_coll;
	struct R_Collbuf *r_coll;

	if(comm->np <= 1) {
		return;
	}

	l_coll = comm->collbuf->l_coll;
	r_coll = comm->collbuf->r_coll;

	allgather = (struct Allgather *) malloc(sizeof(struct Allgather));
	CHECK_ALLOC(allgather);

	dimensions = log_2(comm->np);
	allgather->dimensions = dimensions;

	allgather->peers = (struct Allgather_Peer *) 
		malloc(dimensions*sizeof(struct Allgather_Peer));
	CHECK_ALLOC(allgather->peers);


	allgather->rdma_flag = (char *)((char*)l_coll->buf + ALLGATHER_OFFSET + ALLGATHER_COLL_BUF_SIZE - 
			COLL_RDMA_FLAG_SPACE);
	*((char*)allgather->rdma_flag) = ALLGATHER_RDMA_FLAG;
	allgather->rdma_flag_size = 1;

	for (mask=0x1,i=0; i<dimensions; mask<<=1, i++)	{
		dest = comm->local_rank ^ mask;
		for (j=0; j<3; j++)	{
			allgather->peers[i].done[j] = 
				(char *)l_coll->buf + j*COLL_SINGLE_BUF_SIZE + ALLGATHER_OFFSET + ALLGATHER_BUF_SIZE + i;
			allgather->peers[i].r_done[j] = 
				(char *)r_coll[dest].buf + j*COLL_SINGLE_BUF_SIZE + ALLGATHER_OFFSET + ALLGATHER_BUF_SIZE + i;
#ifdef CH_GEN2
			allgather->peers[i].recv_addr[j] = 
				(char *)l_coll->buf + j*COLL_SINGLE_BUF_SIZE + ALLGATHER_OFFSET + ALLGATHER_BUF_SIZE
				+ COLL_FLAG_SPACE + i*(sizeof(uint32_t)+sizeof(aint_t)+sizeof(char));
			allgather->peers[i].dest_addr[j] = 
				(char *)r_coll[dest].buf + j*COLL_SINGLE_BUF_SIZE + ALLGATHER_OFFSET + ALLGATHER_BUF_SIZE + 
                COLL_FLAG_SPACE	+ i*(sizeof(uint32_t)+sizeof(aint_t)+sizeof(char));
#else
			allgather->peers[i].recv_addr[j] = 
				(char *)l_coll->buf + j*COLL_SINGLE_BUF_SIZE + ALLGATHER_OFFSET + ALLGATHER_BUF_SIZE
				+ COLL_FLAG_SPACE + i*(sizeof(VAPI_rkey_t)+sizeof(aint_t)+sizeof(char));
			allgather->peers[i].dest_addr[j] = 
				(char *)r_coll[dest].buf + j*COLL_SINGLE_BUF_SIZE + ALLGATHER_OFFSET + ALLGATHER_BUF_SIZE + 
                COLL_FLAG_SPACE	+ i*(sizeof(VAPI_rkey_t)+sizeof(aint_t)+sizeof(char));
#endif
		}
	}

	/* This is to alternate b/w the back2back coll buffers */
	allgather->alternate = 0;

	comm->collbuf->allgather = allgather;

#ifdef TIME_ALLGATHER
	comm->collbuf->allgather->send_recv_time = 0;
	comm->collbuf->allgather->reset_time = 0;
	comm->collbuf->allgather->send_complete_time = 0;
#endif

}
#endif
#endif /* VIADEV_RPUT_SUPPORT  */
